Package: glmGamPoi
Type: Package
Title: Fit a Gamma-Poisson Generalized Linear Model
Version: 1.2.0
Authors@R: c(person("Constantin", "Ahlmann-Eltze", email = "artjom31415@googlemail.com", 
                    role = c("aut", "cre"), comment = c(ORCID = "0000-0002-3762-068X")),
             person("Michael", "Love", email="michaelisaiahlove@gmail.com", role = "ctb"))
Description: Fit linear models to overdispersed count data.
     The package can estimate the overdispersion and fit repeated models
     for matrix input. It is designed to handle large input datasets as they
     typically occur in single cell RNA-seq experiments.
License: GPL-3
Encoding: UTF-8
SystemRequirements: C++11
Suggests: testthat (>= 2.1.0), zoo, DESeq2, edgeR, limma, beachmat,
        MASS, statmod, ggplot2, bench, BiocParallel, knitr, rmarkdown,
        BiocStyle, TENxPBMCData, scran
LinkingTo: Rcpp, RcppArmadillo, beachmat
Imports: Rcpp, DelayedMatrixStats, matrixStats, DelayedArray,
        HDF5Array, SummarizedExperiment, methods, stats, utils, splines
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.1.0
URL: https://github.com/const-ae/glmGamPoi
BugReports: https://github.com/const-ae/glmGamPoi/issues
biocViews: Regression, RNASeq, Software, SingleCell
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/glmGamPoi
git_branch: RELEASE_3_12
git_last_commit: a11d461
git_last_commit_date: 2020-10-27
Date/Publication: 2020-10-27
NeedsCompilation: yes
Packaged: 2020-10-28 00:33:35 UTC; biocbuild
Author: Constantin Ahlmann-Eltze [aut, cre]
    (<https://orcid.org/0000-0002-3762-068X>),
  Michael Love [ctb]
Maintainer: Constantin Ahlmann-Eltze <artjom31415@googlemail.com>
