% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_betas.R
\name{estimate_betas_roughly_group_wise}
\alias{estimate_betas_roughly_group_wise}
\title{Make a quick first guess where reasonable beta would be for a set of groups}
\usage{
estimate_betas_roughly_group_wise(Y, offset_matrix, groups)
}
\value{
a matrix with the mean per group for each gene
}
\description{
Make a quick first guess where reasonable beta would be for a set of groups
}
\keyword{internal}
