% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_model.R
\name{format_matrix}
\alias{format_matrix}
\title{Helper to format a matrix nicely}
\usage{
format_matrix(matrix, digits = NULL)
}
\value{
a string
}
\description{
Helper to format a matrix nicely
}
\keyword{internal}
