% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_size_factors.R
\name{estimate_size_factors}
\alias{estimate_size_factors}
\title{Estimate the Size Factors}
\usage{
estimate_size_factors(Y, method, verbose = FALSE)
}
\arguments{
\item{Y}{any matrix-like object (\code{base::matrix()}, \code{DelayedArray}, \code{HDF5Matrix},
\code{Matrix::Matrix()}, etc.)}

\item{method}{one of \code{c("normed_sum", "deconvolution", "poscounts")}}
}
\value{
a vector with one size factor per column of \code{Y}
}
\description{
Estimate the Size Factors
}
\keyword{internal}
