% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_betas.R
\name{estimate_betas_fisher_scoring}
\alias{estimate_betas_fisher_scoring}
\title{Estimate the Betas for Fixed Dispersions}
\usage{
estimate_betas_fisher_scoring(
  Y,
  model_matrix,
  offset_matrix,
  dispersions,
  beta_mat_init
)
}
\value{
a list with two elements
\itemize{
\item \code{Beta} a matrix with one column for each coefficient
\item \code{iterations} the number of iterations
}
}
\description{
Estimate the Betas for Fixed Dispersions
}
\keyword{internal}
