% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pseudobulk.R
\name{vars}
\alias{vars}
\title{Quote grouping variables}
\usage{
vars(...)
}
\arguments{
\item{...}{the quoted expression}
}
\description{
Quote grouping variables
}
\seealso{
ggplot2::vars, dplyr::vars
}
