% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_betas.R
\name{estimate_betas_roughly}
\alias{estimate_betas_roughly}
\title{Make a quick first guess where reasonable beta would be}
\usage{
estimate_betas_roughly(Y, model_matrix, offset_matrix, pseudo_count = 1)
}
\value{
a matrix with one column for each coefficient
}
\description{
Make a quick first guess where reasonable beta would be
}
\keyword{internal}
