% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_betas.R
\name{estimate_betas_group_wise}
\alias{estimate_betas_group_wise}
\title{Estimate the Betas for Fixed Dispersions}
\usage{
estimate_betas_group_wise(
  Y,
  offset_matrix,
  dispersions,
  beta_group_init = NULL,
  beta_mat_init = NULL,
  groups,
  model_matrix
)
}
\value{
a list with three elements
\itemize{
\item \code{Beta} a matrix with one column per group and a row for each gene
\item \code{iterations} the number of iterations from the Newton-Raphson method
\item \code{deviances} the deviance for each gene (sum of the deviance per group)
}
}
\description{
Estimate the Betas for Fixed Dispersions
}
\keyword{internal}
