% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_model.R
\name{print.glmGamPoi}
\alias{print.glmGamPoi}
\alias{format.glmGamPoi}
\alias{summary.glmGamPoi}
\alias{print.summary.glmGamPoi}
\alias{format.summary.glmGamPoi}
\title{Pretty print the result from glm_gp()}
\usage{
\method{print}{glmGamPoi}(x, ...)

\method{format}{glmGamPoi}(x, ...)

\method{summary}{glmGamPoi}(object, ...)

\method{print}{summary.glmGamPoi}(x, ...)

\method{format}{summary.glmGamPoi}(x, ...)
}
\arguments{
\item{x}{the glmGamPoi object}

\item{...}{additional parameters, currently ignored}

\item{object}{the glmGamPoi object that is summarized}
}
\value{
The \code{print()} methods return the object \code{x}. The
\code{format()} method returns a string. The \code{summary()}
method returns an object of class \code{summary.glmGamPoi}.
}
\description{
Pretty print the result from glm_gp()
}
